!        generated by tapenade     (inria, tropics team)
!  tapenade 3.10 (r5363) -  9 sep 2014 09:53
!
! this module contains the source code related to the sa turbulence
! model. it is slightly more modularized than the original which makes
! performing reverse mode ad simplier.
module sa_d
  use constants
  implicit none
  real(kind=realtype) :: cv13, kar2inv, cw36, cb3inv
  real(kind=realtype), dimension(:, :, :), allocatable :: qq
  real(kind=realtype), dimension(:, :, :), pointer :: ddw, ww, ddvt
  real(kind=realtype), dimension(:, :), pointer :: rrlv
  real(kind=realtype), dimension(:, :), pointer :: dd2wall

contains
!  differentiation of sasource in forward (tangent) mode (with options i4 dr8 r8):
!   variations   of useful results: *scratch
!   with respect to varying inputs: timeref *w *rlv *vol *d2wall
!                *si *sj *sk
!   rw status of diff variables: timeref:in *w:in *rlv:in *scratch:out
!                *vol:in *d2wall:in *si:in *sj:in *sk:in
!   plus diff mem management of: w:in rlv:in scratch:in vol:in
!                d2wall:in si:in sj:in sk:in
  subroutine sasource_d()
!
!  source terms.
!  determine the source term and its derivative w.r.t. nutilde
!  for all internal cells of the block.
!  remember that the sa field variable nutilde = w(i,j,k,itu1)
    use blockpointers
    use constants
    use paramturb
    use section
    use inputphysics
    use inputdiscretization, only : approxsa
    use flowvarrefstate
    implicit none
! local parameters
    real(kind=realtype), parameter :: f23=two*third
! local variables.
    integer(kind=inttype) :: i, j, k, nn, ii
    real(kind=realtype) :: fv1, fv2, ft2
    real(kind=realtype) :: fv1d, fv2d, ft2d
    real(kind=realtype) :: ss, sst, nu, dist2inv, chi, chi2, chi3
    real(kind=realtype) :: ssd, sstd, nud, dist2invd, chid, chi2d, chi3d
    real(kind=realtype) :: rr, gg, gg6, termfw, fwsa, term1, term2
    real(kind=realtype) :: rrd, ggd, gg6d, termfwd, fwsad, term1d, &
&   term2d
    real(kind=realtype) :: dfv1, dfv2, dft2, drr, dgg, dfw
    real(kind=realtype) :: uux, uuy, uuz, vvx, vvy, vvz, wwx, wwy, wwz
    real(kind=realtype) :: uuxd, uuyd, uuzd, vvxd, vvyd, vvzd, wwxd, &
&   wwyd, wwzd
    real(kind=realtype) :: div2, fact, sxx, syy, szz, sxy, sxz, syz
    real(kind=realtype) :: div2d, factd, sxxd, syyd, szzd, sxyd, sxzd, &
&   syzd
    real(kind=realtype) :: vortx, vorty, vortz
    real(kind=realtype) :: vortxd, vortyd, vortzd
    real(kind=realtype) :: omegax, omegay, omegaz
    real(kind=realtype) :: omegaxd, omegayd, omegazd
    real(kind=realtype) :: strainmag2, strainprod, vortprod
    real(kind=realtype) :: strainmag2d, strainprodd, vortprodd
    real(kind=realtype), parameter :: xminn=1.e-10_realtype
    intrinsic sqrt
    intrinsic exp
    intrinsic min
    intrinsic max
    real(kind=realtype) :: pwx1
    real(kind=realtype) :: pwx1d
    real(kind=realtype) :: min1
    real(kind=realtype) :: min1d
    real(kind=realtype) :: y1
    real(kind=realtype) :: y1d
! set model constants
    cv13 = rsacv1**3
    kar2inv = one/rsak**2
    cw36 = rsacw3**6
    cb3inv = one/rsacb3
! determine the non-dimensional wheel speed of this block.
    omegaxd = sections(sectionid)%rotrate(1)*timerefd
    omegax = timeref*sections(sectionid)%rotrate(1)
    omegayd = sections(sectionid)%rotrate(2)*timerefd
    omegay = timeref*sections(sectionid)%rotrate(2)
    omegazd = sections(sectionid)%rotrate(3)*timerefd
    omegaz = timeref*sections(sectionid)%rotrate(3)
! create switches to production term depending on the variable that
! should be used
    if (turbprod .eq. katolaunder) then
      print*, 'katolaunder production term not supported for sa'
      stop
    else
      scratchd = 0.0_8
      strainmag2d = 0.0_8
      ssd = 0.0_8
      do k=2,kl
        do j=2,jl
          do i=2,il
! compute the gradient of u in the cell center. use is made
! of the fact that the surrounding normals sum up to zero,
! such that the cell i,j,k does not give a contribution.
! the gradient is scaled by the factor 2*vol.
            uuxd = wd(i+1, j, k, ivx)*si(i, j, k, 1) + w(i+1, j, k, ivx)&
&             *sid(i, j, k, 1) - wd(i-1, j, k, ivx)*si(i-1, j, k, 1) - w&
&             (i-1, j, k, ivx)*sid(i-1, j, k, 1) + wd(i, j+1, k, ivx)*sj&
&             (i, j, k, 1) + w(i, j+1, k, ivx)*sjd(i, j, k, 1) - wd(i, j&
&             -1, k, ivx)*sj(i, j-1, k, 1) - w(i, j-1, k, ivx)*sjd(i, j-&
&             1, k, 1) + wd(i, j, k+1, ivx)*sk(i, j, k, 1) + w(i, j, k+1&
&             , ivx)*skd(i, j, k, 1) - wd(i, j, k-1, ivx)*sk(i, j, k-1, &
&             1) - w(i, j, k-1, ivx)*skd(i, j, k-1, 1)
            uux = w(i+1, j, k, ivx)*si(i, j, k, 1) - w(i-1, j, k, ivx)*&
&             si(i-1, j, k, 1) + w(i, j+1, k, ivx)*sj(i, j, k, 1) - w(i&
&             , j-1, k, ivx)*sj(i, j-1, k, 1) + w(i, j, k+1, ivx)*sk(i, &
&             j, k, 1) - w(i, j, k-1, ivx)*sk(i, j, k-1, 1)
            uuyd = wd(i+1, j, k, ivx)*si(i, j, k, 2) + w(i+1, j, k, ivx)&
&             *sid(i, j, k, 2) - wd(i-1, j, k, ivx)*si(i-1, j, k, 2) - w&
&             (i-1, j, k, ivx)*sid(i-1, j, k, 2) + wd(i, j+1, k, ivx)*sj&
&             (i, j, k, 2) + w(i, j+1, k, ivx)*sjd(i, j, k, 2) - wd(i, j&
&             -1, k, ivx)*sj(i, j-1, k, 2) - w(i, j-1, k, ivx)*sjd(i, j-&
&             1, k, 2) + wd(i, j, k+1, ivx)*sk(i, j, k, 2) + w(i, j, k+1&
&             , ivx)*skd(i, j, k, 2) - wd(i, j, k-1, ivx)*sk(i, j, k-1, &
&             2) - w(i, j, k-1, ivx)*skd(i, j, k-1, 2)
            uuy = w(i+1, j, k, ivx)*si(i, j, k, 2) - w(i-1, j, k, ivx)*&
&             si(i-1, j, k, 2) + w(i, j+1, k, ivx)*sj(i, j, k, 2) - w(i&
&             , j-1, k, ivx)*sj(i, j-1, k, 2) + w(i, j, k+1, ivx)*sk(i, &
&             j, k, 2) - w(i, j, k-1, ivx)*sk(i, j, k-1, 2)
            uuzd = wd(i+1, j, k, ivx)*si(i, j, k, 3) + w(i+1, j, k, ivx)&
&             *sid(i, j, k, 3) - wd(i-1, j, k, ivx)*si(i-1, j, k, 3) - w&
&             (i-1, j, k, ivx)*sid(i-1, j, k, 3) + wd(i, j+1, k, ivx)*sj&
&             (i, j, k, 3) + w(i, j+1, k, ivx)*sjd(i, j, k, 3) - wd(i, j&
&             -1, k, ivx)*sj(i, j-1, k, 3) - w(i, j-1, k, ivx)*sjd(i, j-&
&             1, k, 3) + wd(i, j, k+1, ivx)*sk(i, j, k, 3) + w(i, j, k+1&
&             , ivx)*skd(i, j, k, 3) - wd(i, j, k-1, ivx)*sk(i, j, k-1, &
&             3) - w(i, j, k-1, ivx)*skd(i, j, k-1, 3)
            uuz = w(i+1, j, k, ivx)*si(i, j, k, 3) - w(i-1, j, k, ivx)*&
&             si(i-1, j, k, 3) + w(i, j+1, k, ivx)*sj(i, j, k, 3) - w(i&
&             , j-1, k, ivx)*sj(i, j-1, k, 3) + w(i, j, k+1, ivx)*sk(i, &
&             j, k, 3) - w(i, j, k-1, ivx)*sk(i, j, k-1, 3)
! idem for the gradient of v.
            vvxd = wd(i+1, j, k, ivy)*si(i, j, k, 1) + w(i+1, j, k, ivy)&
&             *sid(i, j, k, 1) - wd(i-1, j, k, ivy)*si(i-1, j, k, 1) - w&
&             (i-1, j, k, ivy)*sid(i-1, j, k, 1) + wd(i, j+1, k, ivy)*sj&
&             (i, j, k, 1) + w(i, j+1, k, ivy)*sjd(i, j, k, 1) - wd(i, j&
&             -1, k, ivy)*sj(i, j-1, k, 1) - w(i, j-1, k, ivy)*sjd(i, j-&
&             1, k, 1) + wd(i, j, k+1, ivy)*sk(i, j, k, 1) + w(i, j, k+1&
&             , ivy)*skd(i, j, k, 1) - wd(i, j, k-1, ivy)*sk(i, j, k-1, &
&             1) - w(i, j, k-1, ivy)*skd(i, j, k-1, 1)
            vvx = w(i+1, j, k, ivy)*si(i, j, k, 1) - w(i-1, j, k, ivy)*&
&             si(i-1, j, k, 1) + w(i, j+1, k, ivy)*sj(i, j, k, 1) - w(i&
&             , j-1, k, ivy)*sj(i, j-1, k, 1) + w(i, j, k+1, ivy)*sk(i, &
&             j, k, 1) - w(i, j, k-1, ivy)*sk(i, j, k-1, 1)
            vvyd = wd(i+1, j, k, ivy)*si(i, j, k, 2) + w(i+1, j, k, ivy)&
&             *sid(i, j, k, 2) - wd(i-1, j, k, ivy)*si(i-1, j, k, 2) - w&
&             (i-1, j, k, ivy)*sid(i-1, j, k, 2) + wd(i, j+1, k, ivy)*sj&
&             (i, j, k, 2) + w(i, j+1, k, ivy)*sjd(i, j, k, 2) - wd(i, j&
&             -1, k, ivy)*sj(i, j-1, k, 2) - w(i, j-1, k, ivy)*sjd(i, j-&
&             1, k, 2) + wd(i, j, k+1, ivy)*sk(i, j, k, 2) + w(i, j, k+1&
&             , ivy)*skd(i, j, k, 2) - wd(i, j, k-1, ivy)*sk(i, j, k-1, &
&             2) - w(i, j, k-1, ivy)*skd(i, j, k-1, 2)
            vvy = w(i+1, j, k, ivy)*si(i, j, k, 2) - w(i-1, j, k, ivy)*&
&             si(i-1, j, k, 2) + w(i, j+1, k, ivy)*sj(i, j, k, 2) - w(i&
&             , j-1, k, ivy)*sj(i, j-1, k, 2) + w(i, j, k+1, ivy)*sk(i, &
&             j, k, 2) - w(i, j, k-1, ivy)*sk(i, j, k-1, 2)
            vvzd = wd(i+1, j, k, ivy)*si(i, j, k, 3) + w(i+1, j, k, ivy)&
&             *sid(i, j, k, 3) - wd(i-1, j, k, ivy)*si(i-1, j, k, 3) - w&
&             (i-1, j, k, ivy)*sid(i-1, j, k, 3) + wd(i, j+1, k, ivy)*sj&
&             (i, j, k, 3) + w(i, j+1, k, ivy)*sjd(i, j, k, 3) - wd(i, j&
&             -1, k, ivy)*sj(i, j-1, k, 3) - w(i, j-1, k, ivy)*sjd(i, j-&
&             1, k, 3) + wd(i, j, k+1, ivy)*sk(i, j, k, 3) + w(i, j, k+1&
&             , ivy)*skd(i, j, k, 3) - wd(i, j, k-1, ivy)*sk(i, j, k-1, &
&             3) - w(i, j, k-1, ivy)*skd(i, j, k-1, 3)
            vvz = w(i+1, j, k, ivy)*si(i, j, k, 3) - w(i-1, j, k, ivy)*&
&             si(i-1, j, k, 3) + w(i, j+1, k, ivy)*sj(i, j, k, 3) - w(i&
&             , j-1, k, ivy)*sj(i, j-1, k, 3) + w(i, j, k+1, ivy)*sk(i, &
&             j, k, 3) - w(i, j, k-1, ivy)*sk(i, j, k-1, 3)
! and for the gradient of w.
            wwxd = wd(i+1, j, k, ivz)*si(i, j, k, 1) + w(i+1, j, k, ivz)&
&             *sid(i, j, k, 1) - wd(i-1, j, k, ivz)*si(i-1, j, k, 1) - w&
&             (i-1, j, k, ivz)*sid(i-1, j, k, 1) + wd(i, j+1, k, ivz)*sj&
&             (i, j, k, 1) + w(i, j+1, k, ivz)*sjd(i, j, k, 1) - wd(i, j&
&             -1, k, ivz)*sj(i, j-1, k, 1) - w(i, j-1, k, ivz)*sjd(i, j-&
&             1, k, 1) + wd(i, j, k+1, ivz)*sk(i, j, k, 1) + w(i, j, k+1&
&             , ivz)*skd(i, j, k, 1) - wd(i, j, k-1, ivz)*sk(i, j, k-1, &
&             1) - w(i, j, k-1, ivz)*skd(i, j, k-1, 1)
            wwx = w(i+1, j, k, ivz)*si(i, j, k, 1) - w(i-1, j, k, ivz)*&
&             si(i-1, j, k, 1) + w(i, j+1, k, ivz)*sj(i, j, k, 1) - w(i&
&             , j-1, k, ivz)*sj(i, j-1, k, 1) + w(i, j, k+1, ivz)*sk(i, &
&             j, k, 1) - w(i, j, k-1, ivz)*sk(i, j, k-1, 1)
            wwyd = wd(i+1, j, k, ivz)*si(i, j, k, 2) + w(i+1, j, k, ivz)&
&             *sid(i, j, k, 2) - wd(i-1, j, k, ivz)*si(i-1, j, k, 2) - w&
&             (i-1, j, k, ivz)*sid(i-1, j, k, 2) + wd(i, j+1, k, ivz)*sj&
&             (i, j, k, 2) + w(i, j+1, k, ivz)*sjd(i, j, k, 2) - wd(i, j&
&             -1, k, ivz)*sj(i, j-1, k, 2) - w(i, j-1, k, ivz)*sjd(i, j-&
&             1, k, 2) + wd(i, j, k+1, ivz)*sk(i, j, k, 2) + w(i, j, k+1&
&             , ivz)*skd(i, j, k, 2) - wd(i, j, k-1, ivz)*sk(i, j, k-1, &
&             2) - w(i, j, k-1, ivz)*skd(i, j, k-1, 2)
            wwy = w(i+1, j, k, ivz)*si(i, j, k, 2) - w(i-1, j, k, ivz)*&
&             si(i-1, j, k, 2) + w(i, j+1, k, ivz)*sj(i, j, k, 2) - w(i&
&             , j-1, k, ivz)*sj(i, j-1, k, 2) + w(i, j, k+1, ivz)*sk(i, &
&             j, k, 2) - w(i, j, k-1, ivz)*sk(i, j, k-1, 2)
            wwzd = wd(i+1, j, k, ivz)*si(i, j, k, 3) + w(i+1, j, k, ivz)&
&             *sid(i, j, k, 3) - wd(i-1, j, k, ivz)*si(i-1, j, k, 3) - w&
&             (i-1, j, k, ivz)*sid(i-1, j, k, 3) + wd(i, j+1, k, ivz)*sj&
&             (i, j, k, 3) + w(i, j+1, k, ivz)*sjd(i, j, k, 3) - wd(i, j&
&             -1, k, ivz)*sj(i, j-1, k, 3) - w(i, j-1, k, ivz)*sjd(i, j-&
&             1, k, 3) + wd(i, j, k+1, ivz)*sk(i, j, k, 3) + w(i, j, k+1&
&             , ivz)*skd(i, j, k, 3) - wd(i, j, k-1, ivz)*sk(i, j, k-1, &
&             3) - w(i, j, k-1, ivz)*skd(i, j, k-1, 3)
            wwz = w(i+1, j, k, ivz)*si(i, j, k, 3) - w(i-1, j, k, ivz)*&
&             si(i-1, j, k, 3) + w(i, j+1, k, ivz)*sj(i, j, k, 3) - w(i&
&             , j-1, k, ivz)*sj(i, j-1, k, 3) + w(i, j, k+1, ivz)*sk(i, &
&             j, k, 3) - w(i, j, k-1, ivz)*sk(i, j, k-1, 3)
! compute the components of the stress tensor.
! the combination of the current scaling of the velocity
! gradients (2*vol) and the definition of the stress tensor,
! leads to the factor 1/(4*vol).
            factd = -(fourth*vold(i, j, k)/vol(i, j, k)**2)
            fact = fourth/vol(i, j, k)
            if (turbprod .eq. strain) then
              sxxd = two*(factd*uux+fact*uuxd)
              sxx = two*fact*uux
              syyd = two*(factd*vvy+fact*vvyd)
              syy = two*fact*vvy
              szzd = two*(factd*wwz+fact*wwzd)
              szz = two*fact*wwz
              sxyd = factd*(uuy+vvx) + fact*(uuyd+vvxd)
              sxy = fact*(uuy+vvx)
              sxzd = factd*(uuz+wwx) + fact*(uuzd+wwxd)
              sxz = fact*(uuz+wwx)
              syzd = factd*(vvz+wwy) + fact*(vvzd+wwyd)
              syz = fact*(vvz+wwy)
! compute 2/3 * divergence of velocity squared
              div2d = f23*2*(sxx+syy+szz)*(sxxd+syyd+szzd)
              div2 = f23*(sxx+syy+szz)**2
! compute strain production term
              strainmag2d = two*(2*sxy*sxyd+2*sxz*sxzd+2*syz*syzd) + 2*&
&               sxx*sxxd + 2*syy*syyd + 2*szz*szzd
              strainmag2 = two*(sxy**2+sxz**2+syz**2) + sxx**2 + syy**2 &
&               + szz**2
              strainprodd = two*strainmag2d - div2d
              strainprod = two*strainmag2 - div2
              if (strainprod .eq. 0.0_8) then
                ssd = 0.0_8
              else
                ssd = strainprodd/(2.0*sqrt(strainprod))
              end if
              ss = sqrt(strainprod)
            else if (turbprod .eq. vorticity) then
! compute the three components of the vorticity vector.
! substract the part coming from the rotating frame.
              vortxd = two*(factd*(wwy-vvz)+fact*(wwyd-vvzd)) - two*&
&               omegaxd
              vortx = two*fact*(wwy-vvz) - two*omegax
              vortyd = two*(factd*(uuz-wwx)+fact*(uuzd-wwxd)) - two*&
&               omegayd
              vorty = two*fact*(uuz-wwx) - two*omegay
              vortzd = two*(factd*(vvx-uuy)+fact*(vvxd-uuyd)) - two*&
&               omegazd
              vortz = two*fact*(vvx-uuy) - two*omegaz
! compute the vorticity production term
              vortprodd = 2*vortx*vortxd + 2*vorty*vortyd + 2*vortz*&
&               vortzd
              vortprod = vortx**2 + vorty**2 + vortz**2
! first take the square root of the production term to
! obtain the correct production term for spalart-allmaras.
! we do this to avoid if statements.
              if (vortprod .eq. 0.0_8) then
                ssd = 0.0_8
              else
                ssd = vortprodd/(2.0*sqrt(vortprod))
              end if
              ss = sqrt(vortprod)
            end if
! compute the laminar kinematic viscosity, the inverse of
! wall distance squared, the ratio chi (ratio of nutilde
! and nu) and the functions fv1 and fv2. the latter corrects
! the production term near a viscous wall.
            nud = (rlvd(i, j, k)*w(i, j, k, irho)-rlv(i, j, k)*wd(i, j, &
&             k, irho))/w(i, j, k, irho)**2
            nu = rlv(i, j, k)/w(i, j, k, irho)
            dist2invd = -(one*2*d2wall(i, j, k)*d2walld(i, j, k)/(d2wall&
&             (i, j, k)**2)**2)
            dist2inv = one/d2wall(i, j, k)**2
            chid = (wd(i, j, k, itu1)*nu-w(i, j, k, itu1)*nud)/nu**2
            chi = w(i, j, k, itu1)/nu
            chi2d = chid*chi + chi*chid
            chi2 = chi*chi
            chi3d = chid*chi2 + chi*chi2d
            chi3 = chi*chi2
            fv1d = (chi3d*(chi3+cv13)-chi3*chi3d)/(chi3+cv13)**2
            fv1 = chi3/(chi3+cv13)
            fv2d = -((chid*(one+chi*fv1)-chi*(chid*fv1+chi*fv1d))/(one+&
&             chi*fv1)**2)
            fv2 = one - chi/(one+chi*fv1)
! the function ft2, which is designed to keep a laminar
! solution laminar. when running in fully turbulent mode
! this function should be set to 0.0.
            if (useft2sa) then
              ft2d = -(rsact3*rsact4*chi2d*exp(-(rsact4*chi2)))
              ft2 = rsact3*exp(-(rsact4*chi2))
            else
              ft2 = zero
              ft2d = 0.0_8
            end if
! correct the production term to account for the influence
! of the wall.
            sstd = ssd + kar2inv*(wd(i, j, k, itu1)*fv2*dist2inv+w(i, j&
&             , k, itu1)*(fv2d*dist2inv+fv2*dist2invd))
            sst = ss + w(i, j, k, itu1)*fv2*kar2inv*dist2inv
! add rotation term (userotationsa defined in inputparams.f90)
            if (userotationsa) then
              if (two*strainmag2 .eq. 0.0_8) then
                y1d = 0.0_8
              else
                y1d = two*strainmag2d/(2.0*sqrt(two*strainmag2))
              end if
              y1 = sqrt(two*strainmag2)
              if (zero .gt. y1) then
                min1d = y1d
                min1 = y1
              else
                min1 = zero
                min1d = 0.0_8
              end if
              sstd = sstd + rsacrot*min1d
              sst = sst + rsacrot*min1
            end if
            if (sst .lt. xminn) then
              sst = xminn
              sstd = 0.0_8
            else
              sst = sst
            end if
! compute the function fw. the argument rr is cut off at 10
! to avoid numerical problems. this is ok, because the
! asymptotical value of fw is then already reached.
            rrd = (kar2inv*(wd(i, j, k, itu1)*dist2inv+w(i, j, k, itu1)*&
&             dist2invd)*sst-w(i, j, k, itu1)*kar2inv*dist2inv*sstd)/sst&
&             **2
            rr = w(i, j, k, itu1)*kar2inv*dist2inv/sst
            if (rr .gt. 10.0_realtype) then
              rr = 10.0_realtype
              rrd = 0.0_8
            else
              rr = rr
            end if
            ggd = rrd + rsacw2*(6*rr**5*rrd-rrd)
            gg = rr + rsacw2*(rr**6-rr)
            gg6d = 6*gg**5*ggd
            gg6 = gg**6
            pwx1d = -((one+cw36)*gg6d/(gg6+cw36)**2)
            pwx1 = (one+cw36)/(gg6+cw36)
            if (pwx1 .gt. 0.0_8 .or. (pwx1 .lt. 0.0_8 .and. sixth .eq. &
&               int(sixth))) then
              termfwd = sixth*pwx1**(sixth-1)*pwx1d
            else if (pwx1 .eq. 0.0_8 .and. sixth .eq. 1.0) then
              termfwd = pwx1d
            else
              termfwd = 0.0_8
            end if
            termfw = pwx1**sixth
            fwsad = ggd*termfw + gg*termfwd
            fwsa = gg*termfw
! compute the source term; some terms are saved for the
! linearization. the source term is stored in dvt.
            if (approxsa) then
              term1 = zero
              term1d = 0.0_8
            else
              term1d = rsacb1*((one-ft2)*ssd-ft2d*ss)
              term1 = rsacb1*(one-ft2)*ss
            end if
            term2d = dist2invd*(kar2inv*rsacb1*((one-ft2)*fv2+ft2)-&
&             rsacw1*fwsa) + dist2inv*(kar2inv*rsacb1*((one-ft2)*fv2d-&
&             ft2d*fv2+ft2d)-rsacw1*fwsad)
            term2 = dist2inv*(kar2inv*rsacb1*((one-ft2)*fv2+ft2)-rsacw1*&
&             fwsa)
            scratchd(i, j, k, idvt) = (term1d+term2d*w(i, j, k, itu1)+&
&             term2*wd(i, j, k, itu1))*w(i, j, k, itu1) + (term1+term2*w&
&             (i, j, k, itu1))*wd(i, j, k, itu1)
            scratch(i, j, k, idvt) = (term1+term2*w(i, j, k, itu1))*w(i&
&             , j, k, itu1)
          end do
        end do
      end do
    end if
  end subroutine sasource_d
  subroutine sasource()
!
!  source terms.
!  determine the source term and its derivative w.r.t. nutilde
!  for all internal cells of the block.
!  remember that the sa field variable nutilde = w(i,j,k,itu1)
    use blockpointers
    use constants
    use paramturb
    use section
    use inputphysics
    use inputdiscretization, only : approxsa
    use flowvarrefstate
    implicit none
! local parameters
    real(kind=realtype), parameter :: f23=two*third
! local variables.
    integer(kind=inttype) :: i, j, k, nn, ii
    real(kind=realtype) :: fv1, fv2, ft2
    real(kind=realtype) :: ss, sst, nu, dist2inv, chi, chi2, chi3
    real(kind=realtype) :: rr, gg, gg6, termfw, fwsa, term1, term2
    real(kind=realtype) :: dfv1, dfv2, dft2, drr, dgg, dfw
    real(kind=realtype) :: uux, uuy, uuz, vvx, vvy, vvz, wwx, wwy, wwz
    real(kind=realtype) :: div2, fact, sxx, syy, szz, sxy, sxz, syz
    real(kind=realtype) :: vortx, vorty, vortz
    real(kind=realtype) :: omegax, omegay, omegaz
    real(kind=realtype) :: strainmag2, strainprod, vortprod
    real(kind=realtype), parameter :: xminn=1.e-10_realtype
    intrinsic sqrt
    intrinsic exp
    intrinsic min
    intrinsic max
    real(kind=realtype) :: pwx1
    real(kind=realtype) :: min1
    real(kind=realtype) :: y1
! set model constants
    cv13 = rsacv1**3
    kar2inv = one/rsak**2
    cw36 = rsacw3**6
    cb3inv = one/rsacb3
! determine the non-dimensional wheel speed of this block.
    omegax = timeref*sections(sectionid)%rotrate(1)
    omegay = timeref*sections(sectionid)%rotrate(2)
    omegaz = timeref*sections(sectionid)%rotrate(3)
! create switches to production term depending on the variable that
! should be used
    if (turbprod .eq. katolaunder) then
      print*, 'katolaunder production term not supported for sa'
      stop
    else
      do k=2,kl
        do j=2,jl
          do i=2,il
! compute the gradient of u in the cell center. use is made
! of the fact that the surrounding normals sum up to zero,
! such that the cell i,j,k does not give a contribution.
! the gradient is scaled by the factor 2*vol.
            uux = w(i+1, j, k, ivx)*si(i, j, k, 1) - w(i-1, j, k, ivx)*&
&             si(i-1, j, k, 1) + w(i, j+1, k, ivx)*sj(i, j, k, 1) - w(i&
&             , j-1, k, ivx)*sj(i, j-1, k, 1) + w(i, j, k+1, ivx)*sk(i, &
&             j, k, 1) - w(i, j, k-1, ivx)*sk(i, j, k-1, 1)
            uuy = w(i+1, j, k, ivx)*si(i, j, k, 2) - w(i-1, j, k, ivx)*&
&             si(i-1, j, k, 2) + w(i, j+1, k, ivx)*sj(i, j, k, 2) - w(i&
&             , j-1, k, ivx)*sj(i, j-1, k, 2) + w(i, j, k+1, ivx)*sk(i, &
&             j, k, 2) - w(i, j, k-1, ivx)*sk(i, j, k-1, 2)
            uuz = w(i+1, j, k, ivx)*si(i, j, k, 3) - w(i-1, j, k, ivx)*&
&             si(i-1, j, k, 3) + w(i, j+1, k, ivx)*sj(i, j, k, 3) - w(i&
&             , j-1, k, ivx)*sj(i, j-1, k, 3) + w(i, j, k+1, ivx)*sk(i, &
&             j, k, 3) - w(i, j, k-1, ivx)*sk(i, j, k-1, 3)
! idem for the gradient of v.
            vvx = w(i+1, j, k, ivy)*si(i, j, k, 1) - w(i-1, j, k, ivy)*&
&             si(i-1, j, k, 1) + w(i, j+1, k, ivy)*sj(i, j, k, 1) - w(i&
&             , j-1, k, ivy)*sj(i, j-1, k, 1) + w(i, j, k+1, ivy)*sk(i, &
&             j, k, 1) - w(i, j, k-1, ivy)*sk(i, j, k-1, 1)
            vvy = w(i+1, j, k, ivy)*si(i, j, k, 2) - w(i-1, j, k, ivy)*&
&             si(i-1, j, k, 2) + w(i, j+1, k, ivy)*sj(i, j, k, 2) - w(i&
&             , j-1, k, ivy)*sj(i, j-1, k, 2) + w(i, j, k+1, ivy)*sk(i, &
&             j, k, 2) - w(i, j, k-1, ivy)*sk(i, j, k-1, 2)
            vvz = w(i+1, j, k, ivy)*si(i, j, k, 3) - w(i-1, j, k, ivy)*&
&             si(i-1, j, k, 3) + w(i, j+1, k, ivy)*sj(i, j, k, 3) - w(i&
&             , j-1, k, ivy)*sj(i, j-1, k, 3) + w(i, j, k+1, ivy)*sk(i, &
&             j, k, 3) - w(i, j, k-1, ivy)*sk(i, j, k-1, 3)
! and for the gradient of w.
            wwx = w(i+1, j, k, ivz)*si(i, j, k, 1) - w(i-1, j, k, ivz)*&
&             si(i-1, j, k, 1) + w(i, j+1, k, ivz)*sj(i, j, k, 1) - w(i&
&             , j-1, k, ivz)*sj(i, j-1, k, 1) + w(i, j, k+1, ivz)*sk(i, &
&             j, k, 1) - w(i, j, k-1, ivz)*sk(i, j, k-1, 1)
            wwy = w(i+1, j, k, ivz)*si(i, j, k, 2) - w(i-1, j, k, ivz)*&
&             si(i-1, j, k, 2) + w(i, j+1, k, ivz)*sj(i, j, k, 2) - w(i&
&             , j-1, k, ivz)*sj(i, j-1, k, 2) + w(i, j, k+1, ivz)*sk(i, &
&             j, k, 2) - w(i, j, k-1, ivz)*sk(i, j, k-1, 2)
            wwz = w(i+1, j, k, ivz)*si(i, j, k, 3) - w(i-1, j, k, ivz)*&
&             si(i-1, j, k, 3) + w(i, j+1, k, ivz)*sj(i, j, k, 3) - w(i&
&             , j-1, k, ivz)*sj(i, j-1, k, 3) + w(i, j, k+1, ivz)*sk(i, &
&             j, k, 3) - w(i, j, k-1, ivz)*sk(i, j, k-1, 3)
! compute the components of the stress tensor.
! the combination of the current scaling of the velocity
! gradients (2*vol) and the definition of the stress tensor,
! leads to the factor 1/(4*vol).
            fact = fourth/vol(i, j, k)
            if (turbprod .eq. strain) then
              sxx = two*fact*uux
              syy = two*fact*vvy
              szz = two*fact*wwz
              sxy = fact*(uuy+vvx)
              sxz = fact*(uuz+wwx)
              syz = fact*(vvz+wwy)
! compute 2/3 * divergence of velocity squared
              div2 = f23*(sxx+syy+szz)**2
! compute strain production term
              strainmag2 = two*(sxy**2+sxz**2+syz**2) + sxx**2 + syy**2 &
&               + szz**2
              strainprod = two*strainmag2 - div2
              ss = sqrt(strainprod)
            else if (turbprod .eq. vorticity) then
! compute the three components of the vorticity vector.
! substract the part coming from the rotating frame.
              vortx = two*fact*(wwy-vvz) - two*omegax
              vorty = two*fact*(uuz-wwx) - two*omegay
              vortz = two*fact*(vvx-uuy) - two*omegaz
! compute the vorticity production term
              vortprod = vortx**2 + vorty**2 + vortz**2
! first take the square root of the production term to
! obtain the correct production term for spalart-allmaras.
! we do this to avoid if statements.
              ss = sqrt(vortprod)
            end if
! compute the laminar kinematic viscosity, the inverse of
! wall distance squared, the ratio chi (ratio of nutilde
! and nu) and the functions fv1 and fv2. the latter corrects
! the production term near a viscous wall.
            nu = rlv(i, j, k)/w(i, j, k, irho)
            dist2inv = one/d2wall(i, j, k)**2
            chi = w(i, j, k, itu1)/nu
            chi2 = chi*chi
            chi3 = chi*chi2
            fv1 = chi3/(chi3+cv13)
            fv2 = one - chi/(one+chi*fv1)
! the function ft2, which is designed to keep a laminar
! solution laminar. when running in fully turbulent mode
! this function should be set to 0.0.
            if (useft2sa) then
              ft2 = rsact3*exp(-(rsact4*chi2))
            else
              ft2 = zero
            end if
! correct the production term to account for the influence
! of the wall.
            sst = ss + w(i, j, k, itu1)*fv2*kar2inv*dist2inv
! add rotation term (userotationsa defined in inputparams.f90)
            if (userotationsa) then
              y1 = sqrt(two*strainmag2)
              if (zero .gt. y1) then
                min1 = y1
              else
                min1 = zero
              end if
              sst = sst + rsacrot*min1
            end if
            if (sst .lt. xminn) then
              sst = xminn
            else
              sst = sst
            end if
! compute the function fw. the argument rr is cut off at 10
! to avoid numerical problems. this is ok, because the
! asymptotical value of fw is then already reached.
            rr = w(i, j, k, itu1)*kar2inv*dist2inv/sst
            if (rr .gt. 10.0_realtype) then
              rr = 10.0_realtype
            else
              rr = rr
            end if
            gg = rr + rsacw2*(rr**6-rr)
            gg6 = gg**6
            pwx1 = (one+cw36)/(gg6+cw36)
            termfw = pwx1**sixth
            fwsa = gg*termfw
! compute the source term; some terms are saved for the
! linearization. the source term is stored in dvt.
            if (approxsa) then
              term1 = zero
            else
              term1 = rsacb1*(one-ft2)*ss
            end if
            term2 = dist2inv*(kar2inv*rsacb1*((one-ft2)*fv2+ft2)-rsacw1*&
&             fwsa)
            scratch(i, j, k, idvt) = (term1+term2*w(i, j, k, itu1))*w(i&
&             , j, k, itu1)
          end do
        end do
      end do
    end if
  end subroutine sasource
!  differentiation of saviscous in forward (tangent) mode (with options i4 dr8 r8):
!   variations   of useful results: *scratch
!   with respect to varying inputs: *w *rlv *scratch *vol *si *sj
!                *sk
!   rw status of diff variables: *w:in *rlv:in *scratch:in-out
!                *vol:in *si:in *sj:in *sk:in
!   plus diff mem management of: w:in rlv:in scratch:in vol:in
!                si:in sj:in sk:in
  subroutine saviscous_d()
!
!  viscous term.
!  determine the viscous contribution to the residual
!  for all internal cells of the block.
    use blockpointers
    use paramturb
    implicit none
! local variables.
    integer(kind=inttype) :: i, j, k, nn, ii
    real(kind=realtype) :: nu
    real(kind=realtype) :: nud
    real(kind=realtype) :: fv1, fv2, ft2
    real(kind=realtype) :: voli, volmi, volpi, xm, ym, zm, xp, yp, zp
    real(kind=realtype) :: volid, volmid, volpid, xmd, ymd, zmd, xpd, &
&   ypd, zpd
    real(kind=realtype) :: xa, ya, za, ttm, ttp, cnud, cam, cap
    real(kind=realtype) :: xad, yad, zad, ttmd, ttpd, cnudd, camd, capd
    real(kind=realtype) :: nutm, nutp, num, nup, cdm, cdp
    real(kind=realtype) :: nutmd, nutpd, numd, nupd, cdmd, cdpd
    real(kind=realtype) :: c1m, c1p, c10, b1, c1, d1, qs
    real(kind=realtype) :: c1md, c1pd, c10d
    intrinsic max
! set model constants
    cv13 = rsacv1**3
    kar2inv = one/rsak**2
    cw36 = rsacw3**6
    cb3inv = one/rsacb3
!
!       viscous terms in k-direction.
!
    do k=2,kl
      do j=2,jl
        do i=2,il
! compute the metrics in zeta-direction, i.e. along the
! line k = constant.
          volid = -(one*vold(i, j, k)/vol(i, j, k)**2)
          voli = one/vol(i, j, k)
          volmid = -(two*(vold(i, j, k)+vold(i, j, k-1))/(vol(i, j, k)+&
&           vol(i, j, k-1))**2)
          volmi = two/(vol(i, j, k)+vol(i, j, k-1))
          volpid = -(two*(vold(i, j, k)+vold(i, j, k+1))/(vol(i, j, k)+&
&           vol(i, j, k+1))**2)
          volpi = two/(vol(i, j, k)+vol(i, j, k+1))
          xmd = skd(i, j, k-1, 1)*volmi + sk(i, j, k-1, 1)*volmid
          xm = sk(i, j, k-1, 1)*volmi
          ymd = skd(i, j, k-1, 2)*volmi + sk(i, j, k-1, 2)*volmid
          ym = sk(i, j, k-1, 2)*volmi
          zmd = skd(i, j, k-1, 3)*volmi + sk(i, j, k-1, 3)*volmid
          zm = sk(i, j, k-1, 3)*volmi
          xpd = skd(i, j, k, 1)*volpi + sk(i, j, k, 1)*volpid
          xp = sk(i, j, k, 1)*volpi
          ypd = skd(i, j, k, 2)*volpi + sk(i, j, k, 2)*volpid
          yp = sk(i, j, k, 2)*volpi
          zpd = skd(i, j, k, 3)*volpi + sk(i, j, k, 3)*volpid
          zp = sk(i, j, k, 3)*volpi
          xad = half*((skd(i, j, k, 1)+skd(i, j, k-1, 1))*voli+(sk(i, j&
&           , k, 1)+sk(i, j, k-1, 1))*volid)
          xa = half*(sk(i, j, k, 1)+sk(i, j, k-1, 1))*voli
          yad = half*((skd(i, j, k, 2)+skd(i, j, k-1, 2))*voli+(sk(i, j&
&           , k, 2)+sk(i, j, k-1, 2))*volid)
          ya = half*(sk(i, j, k, 2)+sk(i, j, k-1, 2))*voli
          zad = half*((skd(i, j, k, 3)+skd(i, j, k-1, 3))*voli+(sk(i, j&
&           , k, 3)+sk(i, j, k-1, 3))*volid)
          za = half*(sk(i, j, k, 3)+sk(i, j, k-1, 3))*voli
          ttmd = xmd*xa + xm*xad + ymd*ya + ym*yad + zmd*za + zm*zad
          ttm = xm*xa + ym*ya + zm*za
          ttpd = xpd*xa + xp*xad + ypd*ya + yp*yad + zpd*za + zp*zad
          ttp = xp*xa + yp*ya + zp*za
! ttm and ttp ~ 1/deltax^2
! computation of the viscous terms in zeta-direction; note
! that cross-derivatives are neglected, i.e. the mesh is
! assumed to be orthogonal.
! furthermore, the grad(nu)**2 has been rewritten as
! div(nu grad(nu)) - nu div(grad nu) to enhance stability.
! the second derivative in zeta-direction is constructed as
! the central difference of the first order derivatives, i.e.
! d^2/dzeta^2 = d/dzeta (d/dzeta k+1/2 - d/dzeta k-1/2).
! in this way the metric can be taken into account.
! compute the diffusion coefficients multiplying the nodes
! k+1, k and k-1 in the second derivative. make sure that
! these coefficients are nonnegative.
          cnudd = -(rsacb2*cb3inv*wd(i, j, k, itu1))
          cnud = -(rsacb2*w(i, j, k, itu1)*cb3inv)
          camd = ttmd*cnud + ttm*cnudd
          cam = ttm*cnud
          capd = ttpd*cnud + ttp*cnudd
          cap = ttp*cnud
! compute nutilde at the faces
          nutmd = half*(wd(i, j, k-1, itu1)+wd(i, j, k, itu1))
          nutm = half*(w(i, j, k-1, itu1)+w(i, j, k, itu1))
          nutpd = half*(wd(i, j, k+1, itu1)+wd(i, j, k, itu1))
          nutp = half*(w(i, j, k+1, itu1)+w(i, j, k, itu1))
! compute nu at the faces
          nud = (rlvd(i, j, k)*w(i, j, k, irho)-rlv(i, j, k)*wd(i, j, k&
&           , irho))/w(i, j, k, irho)**2
          nu = rlv(i, j, k)/w(i, j, k, irho)
          numd = half*((rlvd(i, j, k-1)*w(i, j, k-1, irho)-rlv(i, j, k-1&
&           )*wd(i, j, k-1, irho))/w(i, j, k-1, irho)**2+nud)
          num = half*(rlv(i, j, k-1)/w(i, j, k-1, irho)+nu)
          nupd = half*((rlvd(i, j, k+1)*w(i, j, k+1, irho)-rlv(i, j, k+1&
&           )*wd(i, j, k+1, irho))/w(i, j, k+1, irho)**2+nud)
          nup = half*(rlv(i, j, k+1)/w(i, j, k+1, irho)+nu)
          cdmd = cb3inv*((numd+(one+rsacb2)*nutmd)*ttm+(num+(one+rsacb2)&
&           *nutm)*ttmd)
          cdm = (num+(one+rsacb2)*nutm)*ttm*cb3inv
          cdpd = cb3inv*((nupd+(one+rsacb2)*nutpd)*ttp+(nup+(one+rsacb2)&
&           *nutp)*ttpd)
          cdp = (nup+(one+rsacb2)*nutp)*ttp*cb3inv
          if (cdm + cam .lt. zero) then
            c1m = zero
            c1md = 0.0_8
          else
            c1md = cdmd + camd
            c1m = cdm + cam
          end if
          if (cdp + cap .lt. zero) then
            c1p = zero
            c1pd = 0.0_8
          else
            c1pd = cdpd + capd
            c1p = cdp + cap
          end if
          c10d = c1md + c1pd
          c10 = c1m + c1p
! update the residual for this cell and store the possible
! coefficients for the matrix in b1, c1 and d1.
          scratchd(i, j, k, idvt) = scratchd(i, j, k, idvt) + c1md*w(i, &
&           j, k-1, itu1) + c1m*wd(i, j, k-1, itu1) - c10d*w(i, j, k, &
&           itu1) - c10*wd(i, j, k, itu1) + c1pd*w(i, j, k+1, itu1) + &
&           c1p*wd(i, j, k+1, itu1)
          scratch(i, j, k, idvt) = scratch(i, j, k, idvt) + c1m*w(i, j, &
&           k-1, itu1) - c10*w(i, j, k, itu1) + c1p*w(i, j, k+1, itu1)
        end do
      end do
    end do
!
!       viscous terms in j-direction.
!
    do k=2,kl
      do j=2,jl
        do i=2,il
! compute the metrics in eta-direction, i.e. along the
! line j = constant.
          volid = -(one*vold(i, j, k)/vol(i, j, k)**2)
          voli = one/vol(i, j, k)
          volmid = -(two*(vold(i, j, k)+vold(i, j-1, k))/(vol(i, j, k)+&
&           vol(i, j-1, k))**2)
          volmi = two/(vol(i, j, k)+vol(i, j-1, k))
          volpid = -(two*(vold(i, j, k)+vold(i, j+1, k))/(vol(i, j, k)+&
&           vol(i, j+1, k))**2)
          volpi = two/(vol(i, j, k)+vol(i, j+1, k))
          xmd = sjd(i, j-1, k, 1)*volmi + sj(i, j-1, k, 1)*volmid
          xm = sj(i, j-1, k, 1)*volmi
          ymd = sjd(i, j-1, k, 2)*volmi + sj(i, j-1, k, 2)*volmid
          ym = sj(i, j-1, k, 2)*volmi
          zmd = sjd(i, j-1, k, 3)*volmi + sj(i, j-1, k, 3)*volmid
          zm = sj(i, j-1, k, 3)*volmi
          xpd = sjd(i, j, k, 1)*volpi + sj(i, j, k, 1)*volpid
          xp = sj(i, j, k, 1)*volpi
          ypd = sjd(i, j, k, 2)*volpi + sj(i, j, k, 2)*volpid
          yp = sj(i, j, k, 2)*volpi
          zpd = sjd(i, j, k, 3)*volpi + sj(i, j, k, 3)*volpid
          zp = sj(i, j, k, 3)*volpi
          xad = half*((sjd(i, j, k, 1)+sjd(i, j-1, k, 1))*voli+(sj(i, j&
&           , k, 1)+sj(i, j-1, k, 1))*volid)
          xa = half*(sj(i, j, k, 1)+sj(i, j-1, k, 1))*voli
          yad = half*((sjd(i, j, k, 2)+sjd(i, j-1, k, 2))*voli+(sj(i, j&
&           , k, 2)+sj(i, j-1, k, 2))*volid)
          ya = half*(sj(i, j, k, 2)+sj(i, j-1, k, 2))*voli
          zad = half*((sjd(i, j, k, 3)+sjd(i, j-1, k, 3))*voli+(sj(i, j&
&           , k, 3)+sj(i, j-1, k, 3))*volid)
          za = half*(sj(i, j, k, 3)+sj(i, j-1, k, 3))*voli
          ttmd = xmd*xa + xm*xad + ymd*ya + ym*yad + zmd*za + zm*zad
          ttm = xm*xa + ym*ya + zm*za
          ttpd = xpd*xa + xp*xad + ypd*ya + yp*yad + zpd*za + zp*zad
          ttp = xp*xa + yp*ya + zp*za
! computation of the viscous terms in eta-direction; note
! that cross-derivatives are neglected, i.e. the mesh is
! assumed to be orthogonal.
! furthermore, the grad(nu)**2 has been rewritten as
! div(nu grad(nu)) - nu div(grad nu) to enhance stability.
! the second derivative in eta-direction is constructed as
! the central difference of the first order derivatives, i.e.
! d^2/deta^2 = d/deta (d/deta j+1/2 - d/deta j-1/2).
! in this way the metric can be taken into account.
! compute the diffusion coefficients multiplying the nodes
! j+1, j and j-1 in the second derivative. make sure that
! these coefficients are nonnegative.
          cnudd = -(rsacb2*cb3inv*wd(i, j, k, itu1))
          cnud = -(rsacb2*w(i, j, k, itu1)*cb3inv)
          camd = ttmd*cnud + ttm*cnudd
          cam = ttm*cnud
          capd = ttpd*cnud + ttp*cnudd
          cap = ttp*cnud
          nutmd = half*(wd(i, j-1, k, itu1)+wd(i, j, k, itu1))
          nutm = half*(w(i, j-1, k, itu1)+w(i, j, k, itu1))
          nutpd = half*(wd(i, j+1, k, itu1)+wd(i, j, k, itu1))
          nutp = half*(w(i, j+1, k, itu1)+w(i, j, k, itu1))
          nud = (rlvd(i, j, k)*w(i, j, k, irho)-rlv(i, j, k)*wd(i, j, k&
&           , irho))/w(i, j, k, irho)**2
          nu = rlv(i, j, k)/w(i, j, k, irho)
          numd = half*((rlvd(i, j-1, k)*w(i, j-1, k, irho)-rlv(i, j-1, k&
&           )*wd(i, j-1, k, irho))/w(i, j-1, k, irho)**2+nud)
          num = half*(rlv(i, j-1, k)/w(i, j-1, k, irho)+nu)
          nupd = half*((rlvd(i, j+1, k)*w(i, j+1, k, irho)-rlv(i, j+1, k&
&           )*wd(i, j+1, k, irho))/w(i, j+1, k, irho)**2+nud)
          nup = half*(rlv(i, j+1, k)/w(i, j+1, k, irho)+nu)
          cdmd = cb3inv*((numd+(one+rsacb2)*nutmd)*ttm+(num+(one+rsacb2)&
&           *nutm)*ttmd)
          cdm = (num+(one+rsacb2)*nutm)*ttm*cb3inv
          cdpd = cb3inv*((nupd+(one+rsacb2)*nutpd)*ttp+(nup+(one+rsacb2)&
&           *nutp)*ttpd)
          cdp = (nup+(one+rsacb2)*nutp)*ttp*cb3inv
          if (cdm + cam .lt. zero) then
            c1m = zero
            c1md = 0.0_8
          else
            c1md = cdmd + camd
            c1m = cdm + cam
          end if
          if (cdp + cap .lt. zero) then
            c1p = zero
            c1pd = 0.0_8
          else
            c1pd = cdpd + capd
            c1p = cdp + cap
          end if
          c10d = c1md + c1pd
          c10 = c1m + c1p
! update the residual for this cell and store the possible
! coefficients for the matrix in b1, c1 and d1.
          scratchd(i, j, k, idvt) = scratchd(i, j, k, idvt) + c1md*w(i, &
&           j-1, k, itu1) + c1m*wd(i, j-1, k, itu1) - c10d*w(i, j, k, &
&           itu1) - c10*wd(i, j, k, itu1) + c1pd*w(i, j+1, k, itu1) + &
&           c1p*wd(i, j+1, k, itu1)
          scratch(i, j, k, idvt) = scratch(i, j, k, idvt) + c1m*w(i, j-1&
&           , k, itu1) - c10*w(i, j, k, itu1) + c1p*w(i, j+1, k, itu1)
        end do
      end do
    end do
!
!       viscous terms in i-direction.
!
    do k=2,kl
      do j=2,jl
        do i=2,il
! compute the metrics in xi-direction, i.e. along the
! line i = constant.
          volid = -(one*vold(i, j, k)/vol(i, j, k)**2)
          voli = one/vol(i, j, k)
          volmid = -(two*(vold(i, j, k)+vold(i-1, j, k))/(vol(i, j, k)+&
&           vol(i-1, j, k))**2)
          volmi = two/(vol(i, j, k)+vol(i-1, j, k))
          volpid = -(two*(vold(i, j, k)+vold(i+1, j, k))/(vol(i, j, k)+&
&           vol(i+1, j, k))**2)
          volpi = two/(vol(i, j, k)+vol(i+1, j, k))
          xmd = sid(i-1, j, k, 1)*volmi + si(i-1, j, k, 1)*volmid
          xm = si(i-1, j, k, 1)*volmi
          ymd = sid(i-1, j, k, 2)*volmi + si(i-1, j, k, 2)*volmid
          ym = si(i-1, j, k, 2)*volmi
          zmd = sid(i-1, j, k, 3)*volmi + si(i-1, j, k, 3)*volmid
          zm = si(i-1, j, k, 3)*volmi
          xpd = sid(i, j, k, 1)*volpi + si(i, j, k, 1)*volpid
          xp = si(i, j, k, 1)*volpi
          ypd = sid(i, j, k, 2)*volpi + si(i, j, k, 2)*volpid
          yp = si(i, j, k, 2)*volpi
          zpd = sid(i, j, k, 3)*volpi + si(i, j, k, 3)*volpid
          zp = si(i, j, k, 3)*volpi
          xad = half*((sid(i, j, k, 1)+sid(i-1, j, k, 1))*voli+(si(i, j&
&           , k, 1)+si(i-1, j, k, 1))*volid)
          xa = half*(si(i, j, k, 1)+si(i-1, j, k, 1))*voli
          yad = half*((sid(i, j, k, 2)+sid(i-1, j, k, 2))*voli+(si(i, j&
&           , k, 2)+si(i-1, j, k, 2))*volid)
          ya = half*(si(i, j, k, 2)+si(i-1, j, k, 2))*voli
          zad = half*((sid(i, j, k, 3)+sid(i-1, j, k, 3))*voli+(si(i, j&
&           , k, 3)+si(i-1, j, k, 3))*volid)
          za = half*(si(i, j, k, 3)+si(i-1, j, k, 3))*voli
          ttmd = xmd*xa + xm*xad + ymd*ya + ym*yad + zmd*za + zm*zad
          ttm = xm*xa + ym*ya + zm*za
          ttpd = xpd*xa + xp*xad + ypd*ya + yp*yad + zpd*za + zp*zad
          ttp = xp*xa + yp*ya + zp*za
! computation of the viscous terms in xi-direction; note
! that cross-derivatives are neglected, i.e. the mesh is
! assumed to be orthogonal.
! furthermore, the grad(nu)**2 has been rewritten as
! div(nu grad(nu)) - nu div(grad nu) to enhance stability.
! the second derivative in xi-direction is constructed as
! the central difference of the first order derivatives, i.e.
! d^2/dxi^2 = d/dxi (d/dxi i+1/2 - d/dxi i-1/2).
! in this way the metric can be taken into account.
! compute the diffusion coefficients multiplying the nodes
! i+1, i and i-1 in the second derivative. make sure that
! these coefficients are nonnegative.
          cnudd = -(rsacb2*cb3inv*wd(i, j, k, itu1))
          cnud = -(rsacb2*w(i, j, k, itu1)*cb3inv)
          camd = ttmd*cnud + ttm*cnudd
          cam = ttm*cnud
          capd = ttpd*cnud + ttp*cnudd
          cap = ttp*cnud
          nutmd = half*(wd(i-1, j, k, itu1)+wd(i, j, k, itu1))
          nutm = half*(w(i-1, j, k, itu1)+w(i, j, k, itu1))
          nutpd = half*(wd(i+1, j, k, itu1)+wd(i, j, k, itu1))
          nutp = half*(w(i+1, j, k, itu1)+w(i, j, k, itu1))
          nud = (rlvd(i, j, k)*w(i, j, k, irho)-rlv(i, j, k)*wd(i, j, k&
&           , irho))/w(i, j, k, irho)**2
          nu = rlv(i, j, k)/w(i, j, k, irho)
          numd = half*((rlvd(i-1, j, k)*w(i-1, j, k, irho)-rlv(i-1, j, k&
&           )*wd(i-1, j, k, irho))/w(i-1, j, k, irho)**2+nud)
          num = half*(rlv(i-1, j, k)/w(i-1, j, k, irho)+nu)
          nupd = half*((rlvd(i+1, j, k)*w(i+1, j, k, irho)-rlv(i+1, j, k&
&           )*wd(i+1, j, k, irho))/w(i+1, j, k, irho)**2+nud)
          nup = half*(rlv(i+1, j, k)/w(i+1, j, k, irho)+nu)
          cdmd = cb3inv*((numd+(one+rsacb2)*nutmd)*ttm+(num+(one+rsacb2)&
&           *nutm)*ttmd)
          cdm = (num+(one+rsacb2)*nutm)*ttm*cb3inv
          cdpd = cb3inv*((nupd+(one+rsacb2)*nutpd)*ttp+(nup+(one+rsacb2)&
&           *nutp)*ttpd)
          cdp = (nup+(one+rsacb2)*nutp)*ttp*cb3inv
          if (cdm + cam .lt. zero) then
            c1m = zero
            c1md = 0.0_8
          else
            c1md = cdmd + camd
            c1m = cdm + cam
          end if
          if (cdp + cap .lt. zero) then
            c1p = zero
            c1pd = 0.0_8
          else
            c1pd = cdpd + capd
            c1p = cdp + cap
          end if
          c10d = c1md + c1pd
          c10 = c1m + c1p
! update the residual for this cell and store the possible
! coefficients for the matrix in b1, c1 and d1.
          scratchd(i, j, k, idvt) = scratchd(i, j, k, idvt) + c1md*w(i-1&
&           , j, k, itu1) + c1m*wd(i-1, j, k, itu1) - c10d*w(i, j, k, &
&           itu1) - c10*wd(i, j, k, itu1) + c1pd*w(i+1, j, k, itu1) + &
&           c1p*wd(i+1, j, k, itu1)
          scratch(i, j, k, idvt) = scratch(i, j, k, idvt) + c1m*w(i-1, j&
&           , k, itu1) - c10*w(i, j, k, itu1) + c1p*w(i+1, j, k, itu1)
        end do
      end do
    end do
  end subroutine saviscous_d
  subroutine saviscous()
!
!  viscous term.
!  determine the viscous contribution to the residual
!  for all internal cells of the block.
    use blockpointers
    use paramturb
    implicit none
! local variables.
    integer(kind=inttype) :: i, j, k, nn, ii
    real(kind=realtype) :: nu
    real(kind=realtype) :: fv1, fv2, ft2
    real(kind=realtype) :: voli, volmi, volpi, xm, ym, zm, xp, yp, zp
    real(kind=realtype) :: xa, ya, za, ttm, ttp, cnud, cam, cap
    real(kind=realtype) :: nutm, nutp, num, nup, cdm, cdp
    real(kind=realtype) :: c1m, c1p, c10, b1, c1, d1, qs
    intrinsic max
! set model constants
    cv13 = rsacv1**3
    kar2inv = one/rsak**2
    cw36 = rsacw3**6
    cb3inv = one/rsacb3
!
!       viscous terms in k-direction.
!
    do k=2,kl
      do j=2,jl
        do i=2,il
! compute the metrics in zeta-direction, i.e. along the
! line k = constant.
          voli = one/vol(i, j, k)
          volmi = two/(vol(i, j, k)+vol(i, j, k-1))
          volpi = two/(vol(i, j, k)+vol(i, j, k+1))
          xm = sk(i, j, k-1, 1)*volmi
          ym = sk(i, j, k-1, 2)*volmi
          zm = sk(i, j, k-1, 3)*volmi
          xp = sk(i, j, k, 1)*volpi
          yp = sk(i, j, k, 2)*volpi
          zp = sk(i, j, k, 3)*volpi
          xa = half*(sk(i, j, k, 1)+sk(i, j, k-1, 1))*voli
          ya = half*(sk(i, j, k, 2)+sk(i, j, k-1, 2))*voli
          za = half*(sk(i, j, k, 3)+sk(i, j, k-1, 3))*voli
          ttm = xm*xa + ym*ya + zm*za
          ttp = xp*xa + yp*ya + zp*za
! ttm and ttp ~ 1/deltax^2
! computation of the viscous terms in zeta-direction; note
! that cross-derivatives are neglected, i.e. the mesh is
! assumed to be orthogonal.
! furthermore, the grad(nu)**2 has been rewritten as
! div(nu grad(nu)) - nu div(grad nu) to enhance stability.
! the second derivative in zeta-direction is constructed as
! the central difference of the first order derivatives, i.e.
! d^2/dzeta^2 = d/dzeta (d/dzeta k+1/2 - d/dzeta k-1/2).
! in this way the metric can be taken into account.
! compute the diffusion coefficients multiplying the nodes
! k+1, k and k-1 in the second derivative. make sure that
! these coefficients are nonnegative.
          cnud = -(rsacb2*w(i, j, k, itu1)*cb3inv)
          cam = ttm*cnud
          cap = ttp*cnud
! compute nutilde at the faces
          nutm = half*(w(i, j, k-1, itu1)+w(i, j, k, itu1))
          nutp = half*(w(i, j, k+1, itu1)+w(i, j, k, itu1))
! compute nu at the faces
          nu = rlv(i, j, k)/w(i, j, k, irho)
          num = half*(rlv(i, j, k-1)/w(i, j, k-1, irho)+nu)
          nup = half*(rlv(i, j, k+1)/w(i, j, k+1, irho)+nu)
          cdm = (num+(one+rsacb2)*nutm)*ttm*cb3inv
          cdp = (nup+(one+rsacb2)*nutp)*ttp*cb3inv
          if (cdm + cam .lt. zero) then
            c1m = zero
          else
            c1m = cdm + cam
          end if
          if (cdp + cap .lt. zero) then
            c1p = zero
          else
            c1p = cdp + cap
          end if
          c10 = c1m + c1p
! update the residual for this cell and store the possible
! coefficients for the matrix in b1, c1 and d1.
          scratch(i, j, k, idvt) = scratch(i, j, k, idvt) + c1m*w(i, j, &
&           k-1, itu1) - c10*w(i, j, k, itu1) + c1p*w(i, j, k+1, itu1)
        end do
      end do
    end do
!
!       viscous terms in j-direction.
!
    do k=2,kl
      do j=2,jl
        do i=2,il
! compute the metrics in eta-direction, i.e. along the
! line j = constant.
          voli = one/vol(i, j, k)
          volmi = two/(vol(i, j, k)+vol(i, j-1, k))
          volpi = two/(vol(i, j, k)+vol(i, j+1, k))
          xm = sj(i, j-1, k, 1)*volmi
          ym = sj(i, j-1, k, 2)*volmi
          zm = sj(i, j-1, k, 3)*volmi
          xp = sj(i, j, k, 1)*volpi
          yp = sj(i, j, k, 2)*volpi
          zp = sj(i, j, k, 3)*volpi
          xa = half*(sj(i, j, k, 1)+sj(i, j-1, k, 1))*voli
          ya = half*(sj(i, j, k, 2)+sj(i, j-1, k, 2))*voli
          za = half*(sj(i, j, k, 3)+sj(i, j-1, k, 3))*voli
          ttm = xm*xa + ym*ya + zm*za
          ttp = xp*xa + yp*ya + zp*za
! computation of the viscous terms in eta-direction; note
! that cross-derivatives are neglected, i.e. the mesh is
! assumed to be orthogonal.
! furthermore, the grad(nu)**2 has been rewritten as
! div(nu grad(nu)) - nu div(grad nu) to enhance stability.
! the second derivative in eta-direction is constructed as
! the central difference of the first order derivatives, i.e.
! d^2/deta^2 = d/deta (d/deta j+1/2 - d/deta j-1/2).
! in this way the metric can be taken into account.
! compute the diffusion coefficients multiplying the nodes
! j+1, j and j-1 in the second derivative. make sure that
! these coefficients are nonnegative.
          cnud = -(rsacb2*w(i, j, k, itu1)*cb3inv)
          cam = ttm*cnud
          cap = ttp*cnud
          nutm = half*(w(i, j-1, k, itu1)+w(i, j, k, itu1))
          nutp = half*(w(i, j+1, k, itu1)+w(i, j, k, itu1))
          nu = rlv(i, j, k)/w(i, j, k, irho)
          num = half*(rlv(i, j-1, k)/w(i, j-1, k, irho)+nu)
          nup = half*(rlv(i, j+1, k)/w(i, j+1, k, irho)+nu)
          cdm = (num+(one+rsacb2)*nutm)*ttm*cb3inv
          cdp = (nup+(one+rsacb2)*nutp)*ttp*cb3inv
          if (cdm + cam .lt. zero) then
            c1m = zero
          else
            c1m = cdm + cam
          end if
          if (cdp + cap .lt. zero) then
            c1p = zero
          else
            c1p = cdp + cap
          end if
          c10 = c1m + c1p
! update the residual for this cell and store the possible
! coefficients for the matrix in b1, c1 and d1.
          scratch(i, j, k, idvt) = scratch(i, j, k, idvt) + c1m*w(i, j-1&
&           , k, itu1) - c10*w(i, j, k, itu1) + c1p*w(i, j+1, k, itu1)
        end do
      end do
    end do
!
!       viscous terms in i-direction.
!
    do k=2,kl
      do j=2,jl
        do i=2,il
! compute the metrics in xi-direction, i.e. along the
! line i = constant.
          voli = one/vol(i, j, k)
          volmi = two/(vol(i, j, k)+vol(i-1, j, k))
          volpi = two/(vol(i, j, k)+vol(i+1, j, k))
          xm = si(i-1, j, k, 1)*volmi
          ym = si(i-1, j, k, 2)*volmi
          zm = si(i-1, j, k, 3)*volmi
          xp = si(i, j, k, 1)*volpi
          yp = si(i, j, k, 2)*volpi
          zp = si(i, j, k, 3)*volpi
          xa = half*(si(i, j, k, 1)+si(i-1, j, k, 1))*voli
          ya = half*(si(i, j, k, 2)+si(i-1, j, k, 2))*voli
          za = half*(si(i, j, k, 3)+si(i-1, j, k, 3))*voli
          ttm = xm*xa + ym*ya + zm*za
          ttp = xp*xa + yp*ya + zp*za
! computation of the viscous terms in xi-direction; note
! that cross-derivatives are neglected, i.e. the mesh is
! assumed to be orthogonal.
! furthermore, the grad(nu)**2 has been rewritten as
! div(nu grad(nu)) - nu div(grad nu) to enhance stability.
! the second derivative in xi-direction is constructed as
! the central difference of the first order derivatives, i.e.
! d^2/dxi^2 = d/dxi (d/dxi i+1/2 - d/dxi i-1/2).
! in this way the metric can be taken into account.
! compute the diffusion coefficients multiplying the nodes
! i+1, i and i-1 in the second derivative. make sure that
! these coefficients are nonnegative.
          cnud = -(rsacb2*w(i, j, k, itu1)*cb3inv)
          cam = ttm*cnud
          cap = ttp*cnud
          nutm = half*(w(i-1, j, k, itu1)+w(i, j, k, itu1))
          nutp = half*(w(i+1, j, k, itu1)+w(i, j, k, itu1))
          nu = rlv(i, j, k)/w(i, j, k, irho)
          num = half*(rlv(i-1, j, k)/w(i-1, j, k, irho)+nu)
          nup = half*(rlv(i+1, j, k)/w(i+1, j, k, irho)+nu)
          cdm = (num+(one+rsacb2)*nutm)*ttm*cb3inv
          cdp = (nup+(one+rsacb2)*nutp)*ttp*cb3inv
          if (cdm + cam .lt. zero) then
            c1m = zero
          else
            c1m = cdm + cam
          end if
          if (cdp + cap .lt. zero) then
            c1p = zero
          else
            c1p = cdp + cap
          end if
          c10 = c1m + c1p
! update the residual for this cell and store the possible
! coefficients for the matrix in b1, c1 and d1.
          scratch(i, j, k, idvt) = scratch(i, j, k, idvt) + c1m*w(i-1, j&
&           , k, itu1) - c10*w(i, j, k, itu1) + c1p*w(i+1, j, k, itu1)
        end do
      end do
    end do
  end subroutine saviscous
!  differentiation of saresscale in forward (tangent) mode (with options i4 dr8 r8):
!   variations   of useful results: *dw
!   with respect to varying inputs: *dw *scratch
!   rw status of diff variables: *dw:in-out *scratch:in
!   plus diff mem management of: dw:in scratch:in
  subroutine saresscale_d()
!
!  multiply the residual by the volume and store this in dw; this
! * is done for monitoring reasons only. the multiplication with the
! * volume is present to be consistent with the flow residuals; also
!  the negative value is taken, again to be consistent with the
! * flow equations. also multiply by iblank so that no updates occur
!  in holes or the overset boundary.
    use blockpointers
    implicit none
! local variables
    integer(kind=inttype) :: i, j, k, ii
    real(kind=realtype) :: rblank
    intrinsic real
    intrinsic max
    real(kind=realtype) :: x1
    do k=2,kl
      do j=2,jl
        do i=2,il
          x1 = real(iblank(i, j, k), realtype)
          if (x1 .lt. zero) then
            rblank = zero
          else
            rblank = x1
          end if
          dwd(i, j, k, itu1) = -(volref(i, j, k)*rblank*scratchd(i, j, k&
&           , idvt))
          dw(i, j, k, itu1) = -(volref(i, j, k)*scratch(i, j, k, idvt)*&
&           rblank)
        end do
      end do
    end do
  end subroutine saresscale_d
  subroutine saresscale()
!
!  multiply the residual by the volume and store this in dw; this
! * is done for monitoring reasons only. the multiplication with the
! * volume is present to be consistent with the flow residuals; also
!  the negative value is taken, again to be consistent with the
! * flow equations. also multiply by iblank so that no updates occur
!  in holes or the overset boundary.
    use blockpointers
    implicit none
! local variables
    integer(kind=inttype) :: i, j, k, ii
    real(kind=realtype) :: rblank
    intrinsic real
    intrinsic max
    real(kind=realtype) :: x1
    do k=2,kl
      do j=2,jl
        do i=2,il
          x1 = real(iblank(i, j, k), realtype)
          if (x1 .lt. zero) then
            rblank = zero
          else
            rblank = x1
          end if
          dw(i, j, k, itu1) = -(volref(i, j, k)*scratch(i, j, k, idvt)*&
&           rblank)
        end do
      end do
    end do
  end subroutine saresscale
end module sa_d
